import time
import asyncio
import logging
import inspect
from datetime import datetime, timezone, timedelta
from typing import Type, Tuple, Dict, Callable
from .base import Task, Job, cronable
from .queue import Queue, MemoryQueue
from .worker import Worker, Registry


class Scheduler:
    def __init__(self, queue: Queue = None) -> None:
        self.logger = logging.getLogger(__name__)
        self.registry: Registry = {}
        self.queue = queue or MemoryQueue()
        self.worker = Worker(self.registry, self.queue)
        self.iterations = 0
        self.tick = 60

    def register(self, job: Job) -> None:
        name = getattr(job, 'name', getattr(
            job, '__name__', job.__class__.__name__))
        self.registry[name] = job

    async def setup(self) -> None:
        await self.queue.setup()

    async def work(self) -> None:
        await self.worker.start()

    async def time(self) -> None:
        self.iterations += 1
        while self.iterations:
            self.logger.info(f'Scheduling iteration #{self.iterations}...')
            now = datetime.now(timezone.utc)
            target = (now.replace(microsecond=0)
                      + timedelta(seconds=self.tick))
            delay = (target - now).total_seconds()
            await self.schedule()
            await asyncio.sleep(delay)
            self.iterations += 1

    async def schedule(self) -> None:
        moment = datetime.now(timezone.utc)
        for name, job in self.registry.items():
            frequency = getattr(job, 'frequency', '* * * * *')
            if not cronable(frequency, moment):
                continue

            payload = getattr(job, 'payload', {})
            lane = getattr(job, 'lane', '')
            timeout = getattr(job, 'timeout', 300)
            expired_at = int(datetime.timestamp(moment) + timeout)
            task = Task(job=name, lane=lane,
                        expired_at=expired_at, payload=payload)
            await self.queue.put(task)



from typing import Protocol, List, Mapping, AsyncContextManager


class Connection(Protocol):
    async def fetch(self, query: str, *args, **kwargs) -> List[Mapping]:
        """Fetch the given query records"""


class Connector(Protocol):
    async def get(self, *args, **kwargs) -> Connection:
        """Get a connection from the pool"""



import time
import asyncio
import logging
from typing import Type, Tuple, Dict, Callable
from ..base import Job
from ..queue import Queue


Registry = Dict[str, Job]


class Worker:
    def __init__(self, registry: Registry, queue: Queue) -> None:
        self.logger = logging.getLogger(__name__)
        self.registry = registry
        self.queue = queue
        self.iterations = 0
        self.backoff = 3
        self.retries = 3
        self.sleep = 60
        self.rest = 1

    async def start(self) -> None:
        self.iterations += 1
        while self.iterations:
            self.logger.info(
                f'Work iteration #{self.iterations}...')
            task = await self.queue.pick()
            await self._process(task)
            self.iterations += 1

    def stop(self) -> None:
        self.iterations = 0

    async def _process(self, task) -> None:
        if not task:
            return await asyncio.sleep(self.sleep)

        job = self.registry[task.job]
        retries = getattr(job, 'retries', self.retries)
        backoff = getattr(job, 'backoff', self.backoff)
        try:
            await asyncio.wait_for(job(task), timeout=task.timeout)
            await self.queue.remove(task)
        except Exception:
            self.logger.exception('{task.job}. Task processing failed.')
            task.scheduled_at += (backoff * (2 ** task.attempts))
            task.picked_at = 0
            task.failed_at = time.time()
            task.attempts += 1

            if task.attempts <= retries:
                await self.queue.put(task)
            else:
                await self.queue.remove(task)
                self.logger.info(f'{task.job}. Maximum number of retries '
                                 f'reached: {retries}. Task removed.')

        await asyncio.sleep(self.rest)



from typing import Dict, Protocol
from ..base import Task


class Job(Protocol):
    name: str = ''
    lane: str = ''
    timeout: int = 300
    backoff: int = 3
    retries: int = 3
    frequency: str = '* * * * *'
    payload: dict = {}

    async def __call__(self, task: Task) -> Dict:
        """Job callback to be executed."""



from uuid import uuid4
from time import time


class Task:
    def __init__(self, **attributes) -> None:
        self.id = str(attributes.get('id', str(uuid4())))
        self.created_at = int(attributes.get('created_at', time()))
        self.scheduled_at = int(attributes.get(
            'scheduled_at', self.created_at))
        self.picked_at = int(attributes.get('picked_at', 0))
        self.failed_at = int(attributes.get('failed_at', 0))
        self.timeout = int(attributes.get('timeout', 300))
        self.lane = attributes.get('lane', '')
        self.job = attributes.get('job', '')
        self.attempts = attributes.get('attempts', 0)
        self.payload = attributes.get('payload') or {}



import time
from typing import Dict, Optional
from ...base import Task
from ..queue import Queue


class MemoryQueue(Queue):
    def __init__(self) -> None:
        self.content: Dict[str, Task] = {}
        self.time = time.time

    async def setup(self) -> None:
        self._setup = True

    async def put(self, task: Task) -> None:
        self.content[task.id] = task

    async def pick(self) -> Optional[Task]:
        now = self.time()
        tasks = [task for task in self.content.values()
                 if task.scheduled_at <= now and (not task.picked_at or (
                     task.picked_at + task.timeout <= now))]

        if not tasks:
            return None

        tasks.sort(key=lambda task: task.scheduled_at)
        task = tasks.pop(0)
        task.picked_at = int(time.time())
        return task

    async def remove(self, task: Task) -> None:
        if task.id in self.content:
            del self.content[task.id]



import time
import logging
from typing import Dict
from .base import Task
from .queue import Queue, MemoryQueue


class Planner:
    def __init__(self, queue: Queue = None, timeout: int = 300) -> None:
        self.logger = logging.getLogger(__name__)
        self.queue = queue or MemoryQueue()
        self.timeout = timeout

    async def setup(self) -> None:
        await self.queue.setup()

    async def defer(self, job: str, payload: Dict = None, lane: str = '',
                    delay: int = 0, timeout: int = 0) -> None:
        self.logger.info(f'Deferring job <{job}>...')
        scheduled_at = int(time.time()) + delay
        expired_at = scheduled_at + (timeout or self.timeout)
        task = Task(job=job, lane=lane, scheduled_at=scheduled_at,
                    expired_at=expired_at, payload=payload)
        await self.queue.put(task)



import time
import json
import asyncio
import threading
from json import dumps
from uuid import UUID
from datetime import datetime, timezone
from typing import Mapping, Dict, Optional
from ...base import Task
from ..queue import Queue
from .connector import Connector
from .migrations import (
    migrate, TASKS_SCHEMA, TASKS_TABLE)


class SqlQueue(Queue):
    def __init__(self, connector: Connector) -> None:
        self.connector = connector
        self.table = TASKS_TABLE
        self.schema = TASKS_SCHEMA

    async def setup(self):
        await migrate(self.connector)

    async def put(self, task: Task) -> None:
        query = """
        INSERT INTO public.__tasks__ (
            id, created_at, scheduled_at, picked_at, failed_at,
            timeout, lane, job, attempts, payload
        ) VALUES (
            $1, $2, $3, $4, $5, $6, $7, $8, $9, $10
        ) ON CONFLICT (id) DO UPDATE SET (
            created_at, scheduled_at, picked_at, failed_at,
            timeout, lane, job, attempts, payload
        ) = (
            EXCLUDED.created_at, EXCLUDED.scheduled_at,
            EXCLUDED.picked_at, EXCLUDED.failed_at, EXCLUDED.timeout,
            EXCLUDED.lane, EXCLUDED.job, EXCLUDED.attempts,
            EXCLUDED.payload
        )
        RETURNING *
        """

        connection = await self.connector.get()
        parameters = [
            task.id, datetime.fromtimestamp(task.created_at, timezone.utc),
            datetime.fromtimestamp(task.scheduled_at, timezone.utc),
            datetime.fromtimestamp(task.picked_at, timezone.utc),
            datetime.fromtimestamp(task.failed_at, timezone.utc),
            task.timeout, task.lane, task.job, task.attempts,
            dumps(task.payload)]

        await connection.fetch(query, *parameters)

    async def pick(self) -> Optional[Task]:
        query = f"""
        UPDATE public.__tasks__
        SET picked_at = NOW()::timestamptz
        WHERE id = (
            SELECT id FROM public.__tasks__
            WHERE scheduled_at <= NOW()::timestamptz
            AND (picked_at = 'epoch' OR (picked_at
            + interval '1 second' * timeout) <= NOW()::timestamptz)
            ORDER BY scheduled_at
            FOR UPDATE SKIP LOCKED
            LIMIT 1
        )
        RETURNING *
        """

        connection = await self.connector.get()
        result = await connection.fetch(query)

        if not result:
            return None

        record = dict(result[0])
        record['id'] = str(record['id'])
        record['created_at'] = int(datetime.timestamp(
            record['created_at']))
        record['scheduled_at'] = int(datetime.timestamp(
            record['scheduled_at']))
        record['picked_at'] = int(datetime.timestamp(
            record['picked_at']))
        record['failed_at'] = int(datetime.timestamp(
            record['failed_at']))
        record['payload'] = json.loads(
            str(record['payload']))

        return Task(**record)

    async def remove(self, task: Task) -> None:
        query = f"""
        DELETE FROM public.__tasks__
        WHERE id = $1
        """

        connection = await self.connector.get()

        parameters = [task.id]

        await connection.fetch(query, *parameters)



from .connector import Connector, Connection


VERSION_KEY = 'SCHEDULARK_DBVERSION'
SETTINGS_TABLE = '__settings__'
TASKS_SCHEMA = 'public'
TASKS_TABLE = '__tasks__'


async def migrate(connector: Connector) -> None:
    connection = await connector.get()
    migrations = {key.replace('migration_', ''): value for key, value
                  in globals().items() if key.startswith('migration')}

    current = await get_version(connection)
    for target, migration in migrations.items():
        await apply_migration(current, target, migration, connection)


async def apply_migration(current, target, migration, connection) -> None:
    if current >= target:  # pragma: no cover
        return None
    await migration(connection)
    table = f'{TASKS_SCHEMA}.{SETTINGS_TABLE}'
    await connection.fetch(f"""
    INSERT INTO {table} (key, value)
        VALUES ($1, $2)
        ON CONFLICT (key)
        DO UPDATE SET value = EXCLUDED.value
    """, VERSION_KEY, target)


async def get_version(connection: Connection) -> str:
    table = f'{TASKS_SCHEMA}.{SETTINGS_TABLE}'
    await connection.fetch(
        f'CREATE TABLE IF NOT EXISTS {table} (key TEXT UNIQUE, value TEXT)')
    result = await connection.fetch(
        f'SELECT key, value FROM {table} WHERE key = $1', VERSION_KEY)
    row = next(iter(result), {'value': '000'})
    return row['value']


async def migration_001(connection: Connection):
    table = f'{TASKS_SCHEMA}.{TASKS_TABLE}'
    await connection.fetch(f"""
    CREATE TABLE IF NOT EXISTS {table} (
        id UUID PRIMARY KEY,
        created_at TIMESTAMPTZ,
        scheduled_at TIMESTAMPTZ,
        picked_at TIMESTAMPTZ,
        failed_at TIMESTAMPTZ,
        timeout INTEGER,
        lane TEXT,
        job TEXT,
        attempts INTEGER,
        payload JSONB)
    """)



import os
import logging
from .base import Task, Job
from .planner import Planner
from .scheduler import Scheduler
from .queue import Queue, MemoryQueue, SqlQueue, JsonQueue


__version__ = '0.1.7'


logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))



import os
import time
import json
import fcntl
from pathlib import Path
from typing import Dict, Optional
from contextlib import contextmanager
from ...base import Task
from ..queue import Queue


class JsonQueue(Queue):
    def __init__(self, path: str = '') -> None:
        self.path = path or 'tasks.json'
        self.time = time.time

    async def setup(self) -> None:
        path = Path(self.path)
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            with locked_open(self.path, 'w') as f:
                f.write("{}")

    async def put(self, task: Task) -> None:
        if not os.path.exists(self.path):
            await self.setup()

        content: Dict = {}
        with locked_open(self.path, 'r+') as f:
            content.update(json.loads(f.read()))
            content[task.id] = vars(task)
            f.seek(f.truncate(0))
            f.write(json.dumps(content, indent=2))

    async def pick(self) -> Optional[Task]:
        if not os.path.exists(self.path):
            return None

        with locked_open(self.path, 'r+') as f:
            content: Dict = json.loads(f.read())

            now = self.time()
            tasks = [task for task in content.values()
                     if task['scheduled_at'] <= now and (
                         not task['picked_at'] or (
                             task['picked_at']
                             + task['timeout'] <= now))]

            if not tasks:
                return None

            tasks.sort(key=lambda task: task['scheduled_at'])
            task = tasks.pop(0)
            task['picked_at'] = int(time.time())
            content[task['id']] = task

            f.seek(f.truncate(0))
            f.write(json.dumps(content, indent=2))

            return Task(**task)

    async def remove(self, task: Task) -> None:
        if not os.path.exists(self.path):
            return

        with locked_open(self.path, 'r+') as f:
            content: Dict = json.loads(f.read())

            if task.id in content:
                del content[task.id]

            f.seek(f.truncate(0))
            f.write(json.dumps(content, indent=2))


@contextmanager
def locked_open(filename, mode='r'):
    with open(filename, mode) as file:
        fcntl.flock(file, fcntl.LOCK_EX)
        yield file
        fcntl.flock(file, fcntl.LOCK_UN)



from datetime import datetime, timezone


def cronable(pattern: str, moment: datetime = None) -> bool:
    """Check if a datetime moment matches the given cron pattern.
    .--------------- minute (0-59)
    |  .------------ hour (0-23)
    |  |  .--------- month_day (1-31)
    |  |  |  .------ month (1-12)
    |  |  |  |  .--- week_day (1-7; sunday=7)
    |  |  |  |  |
    *  *  *  *  *  cron command pattern
    """
    if not pattern:
        return False

    valid_characters = '0123456789 */!'
    if any(char not in valid_characters for char in pattern):
        raise ValueError(
            f'Invalid cron pattern. Use "{valid_characters}" only.')

    moment = moment or datetime.now(timezone.utc)

    second = '*'
    minute, hour, month_day, month, week_day = pattern.split()
    if '!' in minute:
        symbol, value = minute.split("!")
        second, minute = (f'*/{value}' if symbol else value, '*')

    return (_check_field(second, moment.second)
            and _check_field(minute, moment.minute)
            and _check_field(hour, moment.hour)
            and _check_field(month_day, moment.day)
            and _check_field(month, moment.month)
            and _check_field(week_day, moment.isoweekday()))


def _check_field(field: str, value: int) -> bool:
    if field == '*':
        return True

    if '/' in field:
        _, interval = field.split('/')
        return value % int(interval) == 0

    return int(field) == value



from typing import Optional
from abc import ABC, abstractmethod
from ..base import Task


class Queue(ABC):

    async def setup(self) -> None:
        """Optional setup procedures"""

    @abstractmethod
    async def put(self, task: Task) -> None:
        """Put method to be implemented"""

    @abstractmethod
    async def pick(self) -> Optional[Task]:
        """Retain method to be implemented"""

    @abstractmethod
    async def remove(self, task: Task) -> None:
        """Remove method to be implemented"""
